<?php
namespace Ma;

use Ma\Container;
use Ma\Web\Router;
use Ma\Service\Logger;
use Ma\Service\Tpl;

/**
 * @package         Ma
 * @class           App
 * @extends         \Ma\Container
 * @author          Mahdi Abu Yasmine
 * @date            2019-10-15 16:01:18 CET
 * @copyright       meta-tech.academy
 */
abstract class App extends Container
{
    protected string    $name;
    protected array     $services;

    /**
     * @constructor
     * @public
     * @param       string      $name
     * @param       array       $config
     */
    public function __construct(string $name = 'app', array $config = [])
    {
        $this->name     = $name;
        $this->services = [];
        $this->baseInit($config);
    }

    /**
     * @public
     * @return      string
     */
    public function getName():string
    {
        return $this->name;
    }

    /**
     * @public
     * @return      array
     */
    public function availableServices(): array
    {
        return array_keys($this->services);
    }

    /**
     * @public
     * @param       string      $name
     * @return      bool
     */
    public function serviceExists(string $name): bool
    {
        return in_array($name, $this->availableServices());
    }

    /**
     * @public
     * @param       string  $name
     * @param       mixed   $provider
     * @return      bool
     */
    public function registerService(string $name, $provider): bool
    {
        if (($done = !$this->serviceExists($name))) {
            $this->services[$name] = $provider;
        }
        return $done;
    }

    /**
     * @public
     * @param       string  $prefix
     * @param       array   $definition
     * @return      bool
     */
    public function registerServices(string $prefix, array $definition): bool
    {
        $done = true;
        foreach($definition as $name => $obj) {
            $done = $done && $this->registerService("$prefix$name", $obj);
        }
        return $done;
    }

    /**
     * @abstract
     * @protected
     * @param       array       $config
     */
    protected function baseInit(array $config = [])
    {
        $this->services['config'] = $config;
        $this->services['log']    = function(string $name) use (&$logger) {
            if (is_null($logger)) {
                $logger = new Logger(@$this->services['config']['log']);
            }
            return $logger->get($name);
        };
        $this->services['router'] = new Router();
        $this->services['tpl']    = function(?string $name) use (&$tpl) {
            if (is_null($tpl)) {
                $tpl = new Tpl($this->services['config']['tpl']);
            }
            return $tpl->get($name);
        };
        $this->init($config);
    }

    /**
     * @abstract
     * @public
     * @param       array       $config
     */
    abstract public function init(array $config);

    /**
     * @abstract
     * @public
     */
    abstract public function routing();

    /**
     * @method      __get
     * @public
     * @param       string      $property
     * @return      mixed
     */
    public function __get(string $property)
    {
        return $this->serviceExists($property) ? $this->services[$property] : null;
    }

    /**
     * @public
     * @static
     * @param      ?string      $name
     * @param       array       $params
     * @return      mixed
     */
    public static function get(?string $name, array $params = [])
    {
        global $core;
        $unit = empty($name) ? $core : $core[$name];
        return is_callable($unit) ? call_user_func_array($unit, $params) : $unit;
    }

    /**
     * @public
     */
    public function run()
    {
        $this->services['router']->resolve($_SERVER["REDIRECT_URL"]);
    }

    /**
     * @public
     */
    public function boot()
    {
        $this->routing();
    }

    /**
     * @public
     * @param       string  $path
     */
    public function redirect(string $path)
    {
        $this->endWith(function() {
            header("Location: $path");
            die();
        });
    }

    /**
     * @protected
     * @param       closure $fn
     */
    protected function endWith(closure $fn)
    {
        $s = session_commit();
        $fn();
    }

    /**
     * @public
     * @param       array   $data
     * @param       int     $code
     */
    public function json(array $data = [], int $code=200)
    {
        $this->endWith(function() {
            header("Content-Type: application/json; charset=utf-8");
            try {
                $json = json_encode($data);
            }
            catch(\Exception $e) {
                $error = "Server Error, cannot json encode data";
                $json  = json_encode(compact('error'));
                $this->services['log']('main')->error($error, compact($data));
                $code  = 500;
            }
            http_response_code($code);
            echo $json;
            die();
        });
    }
}
